//
//  Copyright (C) 2004-2006  Autodesk, Inc.
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of version 2.1 of the GNU Lesser
//  General Public License as published by the Free Software Foundation.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#ifndef _FSDPARSER_H
#define _FSDPARSER_H

#include "../MdfModel/FeatureSource.h"

typedef std::ostream MdfStream;

#pragma warning(push)
// Disable C4267 of XmlScanner about size_t assigned to unsigned int
#pragma warning(disable:4267)

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/sax2/SAX2XMLReader.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>
#include <xercesc/sax2/DefaultHandler.hpp>
#include <xercesc/framework/XMLFormatter.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include "SAX2ElementHandler.h"

#pragma warning(pop)

using namespace XERCES_CPP_NAMESPACE;
using namespace MDFMODEL_NAMESPACE;

BEGIN_NAMESPACE_MDFPARSER

/* FSDSAX2Parser is the class for parsing FSD. It holds the actual Xerces SAX2
 * Parser and is also the handler for the events generated by the parser. It
 * holds the main references to the feature source that will be created by the
 * parse. It handles both reading and writing through public functions.
 */
class MDFPARSER_API FSDSAX2Parser : public DefaultHandler
{
    protected:
        // The objects to be generated by the parse.
        FeatureSource *m_pFeatureSource;

        // Succeeded is true if the parse has succeeded. As of now,
        // there are very loose constraints on this boolean.
        bool succeeded;
        
        SAX2XMLReader          *m_Parser;
        // handlerStack is a stack of SAX2ElementHandlers (see the
        // typedef in SAX2ElementHandler.h). When the parser receives
        // events they will be routed to the SAX2ElementHandler on the top
        // of the handlerStack, which will be an instance of one of the IO classes,
        // who all implement the SAX2ElementHandler interface.
        HandlerStack* m_HandlerStack;
        
        // Unfortunately, the SAX2 Parser doesn't necessarily send all the
        // characters that appear inside a tag in one call of characters()-
        // for example with escape characters it separates them into separate calls.
        // Therefore, m_strbuffer is used to accumulate the characters, openTag and
        // prevOpenTag are used to determine if we're still reading characters in the 
        // same open tag, and tagOpen and charsInTag are used in the case when you have 
        // no characters in between a tag (i.e. <Metadata></Metadata>) to make sure
        // that whitespace or line returns don't get assigned to it instead.
        std::wstring m_strbuffer;
        
        std::wstring m_openTag;
        std::wstring m_prevOpenTag;
        MdfString m_strParserError;
        bool m_tagOpen;
        bool m_charsInTag;

        void WriteDefinition(FeatureSource *pFeatureSource, MdfStream &fd);

        // error handling override
        virtual void error(const SAXException &exc);

    public:
        FSDSAX2Parser();
        ~FSDSAX2Parser();
        void Flush();
        void Initialize();
        void DisableTabs();
        void ParseFile(std::string str);
        void ParseFile(char* str);
        void ParseString(const char* str, unsigned int numBytes);
        void WriteToFile(std::string name, FeatureSource *pFeatureSource);

        std::string SerializeToXML(FeatureSource *pFeatureSource);

        FeatureSource* DetachFeatureSource();

        // Occurs when an XML start tag is encountered. 'localname' is the text in
        // the tag to be paid attention to in our schema.
        void startElement(const   XMLCh* const    uri,
                          const   XMLCh* const    localname,
                          const   XMLCh* const    qname,
                          const   Attributes&     attributes);

        // Occurs when an XML closing tag is encountered. 'localname' is the text in
        // the tag to be paid attention to in our schema.
        void endElement(const XMLCh* const uri,
                        const XMLCh* const localname,
                        const XMLCh* const qname);

        // Occurs when the characters in between XML tags is encountered.
        void characters(const XMLCh* const chars, const unsigned int length);

        // get error info
        bool GetSucceeded() const
        {
            return this->succeeded;
        }

        static FeatureSource* CreateClone(FeatureSource *pFeatureSource);
        const MdfString& GetErrorMessage();
};    

END_NAMESPACE_MDFPARSER
#endif // _FSDPARSER_H
